#version 330
//quelle https://developer.nvidia.com/sites/default/files/akamai/gamedev/docs/PCSS_Integration.pdf
//#define BLOCKER_SEARCH_NUM_SAMPLES 16
//#define PCF_NUM_SAMPLES 16
//#define NEAR_PLANE 9.5
//#define LIGHT_WORLD_SIZE 0.5
//#define LIGHT_FRUSTUM_WIDTH 3.75
//#define LIGHT_SIZE_UV (LIGHT_WORLD_SIZE / LIGHT_FRUSTUM_WIDTH)

in vec3 N;
in vec3 V;
in vec3 P;
in vec2 fUVs;
in mat3 TBN;
//TEST
in float flip;

uniform float ambientFactor;
uniform float shininess;
uniform int smType;
uniform int nmType;

uniform sampler2D diffuseTexture;
uniform sampler2D normalTexture;

uniform float opacity;

uniform int wireframe;

uniform float damage;

out vec4 colorOut;

//per spotlight
in vec3 L_0;
in vec4 shadowTexCoord_0;
uniform vec3 spotDir_0;
uniform float spotCutOff_0;
uniform float I_0;
uniform sampler2D shadowMap_0;

in vec3 L_1;
in vec4 shadowTexCoord_1;
uniform vec3 spotDir_1;
uniform float spotCutOff_1;
uniform float I_1;
uniform sampler2D shadowMap_1;


//VSM
float linstep(float low, float high, float v){
	return clamp((v - low)/(high - low), 0.0, 1.0);
}

float VSM(sampler2D map, vec3 coords){
	vec2 moments = texture2D(map, coords.xy).xy; //E(x) und E(x^2)
	float compare = coords.z;
	//voll beleuchtet?
	if(moments.x >= compare){
		return 1.0;
	}

	float near = 0.01;
	float far = 20.0;
	compare = (2 * near) / (far + near - compare * (far - near));
	float p = smoothstep(compare - 0.005, compare, moments.x); //lower, upper, value
	float variance = moments.y - (moments.x * moments.x);
	variance = clamp(variance, 0.0, 1.0);

	float d = compare - moments.x;
	float p_max = variance / (variance + d * d);
	p_max = linstep(0.3, 1.0, p_max); //lightbleeding fix
	return clamp(max(p,p_max), 0.0,1.0);
	
}

void main() {
	vec3 L;
	vec3 spotDir;
	float spotCutOff;
	float I;



	if(wireframe == 1){
		colorOut = vec4(1.0,1.0,1.0,1.0);
	} else {

		vec3 specularColor = vec3(1.0, 0.8, 0.5); //vec3(1.0,0.0,0.0);
		vec3 textureColor = texture(diffuseTexture, fUVs).rgb;		
		//textureColor = vec3(0.8);
		vec3 ambient = ambientFactor * textureColor + (1-ambientFactor) * vec3(0.0,0.0,0.0);
		vec3 textureNormal = normalize(TBN * (texture2D(normalTexture, fUVs).rgb * 2.0 - 1.0)); //TBN * 
		vec3 colorTotal = vec3(0.0);
		vec3 normal;
		if(nmType == -1) normal = normalize(N);
		else {
			normal = textureNormal;
			//normal.x *= flip;
		}

		//normal = normalize(N);
		vec3 view = normalize(V);

		//TODO auerhalb definieren -> struct?
		float constantAttenuation = 0.5;
		float linearAttenuation = 0.05;
		float quadraticAttenuation = 0.05;
		float spotExponent = 0.0;

		for(int i=0; i<2; i++){
			if(i == 0){ L = L_0; spotDir = spotDir_0; spotCutOff = spotCutOff_0; I = I_0; }
			if(i == 1){ L = L_1; spotDir = spotDir_1; spotCutOff = spotCutOff_1; I = I_1; }


			vec4 shadowTexCoord;
			if(i == 0){ shadowTexCoord = shadowTexCoord_0; }
			if(i == 1){ shadowTexCoord = shadowTexCoord_1; }

			float shadow = 0.0;
			vec3 coords = shadowTexCoord.xyz/shadowTexCoord.w;
			bool outsideShadowMap = shadowTexCoord.w <= 0.0
									|| (coords.x < 0 || coords.y < 0)
									|| (coords.x >= 1 || coords.y >= 1);
			if(!outsideShadowMap && i == 0) shadow += VSM(shadowMap_0, coords);
			else if(!outsideShadowMap && i == 1) shadow += VSM(shadowMap_1, coords);
			else shadow += 0.0;


			float lightDistance = length(L);
			vec3 lightDirection = normalize(L);
			//if(nmType != -1) lightDirection = normalize(TBN * L);

			float attenuation = 1.0 / (constantAttenuation
								+ linearAttenuation * lightDistance
								+ quadraticAttenuation * (lightDistance * lightDistance));
			float cosSL = max(0.0, dot(normalize(spotDir), lightDirection));
			if(cosSL < spotCutOff) {
				attenuation = 0.0;
			} else {
				attenuation = attenuation * pow(cosSL, spotExponent);
			}
			
			float cosNL = max(0.0, dot(normal, lightDirection));
			vec3 diffuse = attenuation * I * textureColor * cosNL;
		
			vec3 H = lightDirection + view;
			H = normalize(H); 
			float cosNH = max(0.0, dot(H, normal));
			vec3 specular;
			//if(cosNL < 0.0){
			//	//specular = vec3(0.0,0.0,0.0);
			//} else {
			//	specular = attenuation * I * specularColor * pow(cosNH, shininess);
			//}
			
			specular = attenuation * I * specularColor * pow(cosNH, 10);

			float spotFactor = dot(normalize(L), normalize(spotDir));

			if(spotFactor > spotCutOff){
				spotFactor = 1.0 - (1.0 - spotFactor) * 1.0/(1.0 - spotCutOff);
				colorTotal += spotFactor * shadow * (diffuse + specular) + ambient;
			} else
				colorTotal += ambient;
		}
		colorTotal /= 2.0;

		if(opacity<1.0)	colorOut = vec4(colorTotal, opacity);
		else colorOut = vec4(colorTotal, 1.0);


		//debug
		vec4 debug_normal = vec4(normal, 1.0);
		//colorOut = (debug_normal + 1.0)/2.0;
	}
}
